//---------------------------------------------------------------------------
//
//                      WWW.NATUREWIZARD.COM
//                      ~~~~~~~~~~~~~~~~~~~~
//
//              Author      : Dr. Heinz Seyringer
//              Project     : Naturewizard Tutorial 2
//              Description : This projects illustrates the physical models
//                            behind the simulation of waves.
//
//---------------------------------------------------------------------------
#include <vcl.h>
#include <math.h>
#pragma hdrstop

#include "naturewizard.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TFormMain *FormMain;

typedef struct
{
   double x,y,z;
} VERTEX3D;

typedef struct
{
   double direction;
   double dirX;
   double dirY;
   double lambda;                 // wavelength
   double k;                      // wave number: k=lamda/(2*Pi)
   double omega;                  // angular frequency: omega=2*Pi*f
   double freq;                   // frequency
   double periode;                // periode of the wave
   double amplitude;              // amplitude of the wave
   double phase;                  // phase of the wave
} WAVE;

int nWaves=1;
WAVE wave[256];                   // determines maximum number of waves
const GLfloat GridSize = 5.0f;    // GridSize of the ocean
const float Pi=3.1415926535898;
const float g=9.80665;            // standard gravitational acceleration
const float alpha=0.0081;         // coefficient for Pierson-Moskowitz formula
GLdouble fPeak;                   // frequency peak
GLdouble WindSpeed=10.0;          // wind speed in meters per second
GLdouble WindDirection=0.0;       // direction of the wind
GLint maxwave=15;                 // maximum number of waves used in simulation
GLint WaterX=128;                 // grid resolution in x direction
GLint WaterY=128;                 // grid resolution in y direction

double *WaterHeight = new double [WaterX*WaterY];     // elongation of the wave
VERTEX3D *WaterNormal = new VERTEX3D [WaterX*WaterY]; // normal vector to the water surface

//---------------------------------------------------------------------------
double Gamma (double x)
{
   x=x-1;                                        // change from Gamma function to faculty: G(x)=(x-1)!
   double faculty=1.0f;                          // initialize faculty function
   double ulim=x;                                // define "upper limit" for faculty
   if (floor(x)==x) ulim=x+1.0;                  // correct if x is integer
   for(int i=1;i<ulim;i++)
      faculty*=i;                                // calculate faculty
   double fact=floor(x)*x+1-floor(x)*floor(x);   // calculate last multiplication factor
   faculty*=fact;
   return faculty;
}
//---------------------------------------------------------------------------
double MitsuyasuDistribution (double f, double theta)
{
   double s,D;
   double ang=cos(theta/2);
   ang *=ang;                   // calculate square of cos(theta/2)
   if (f>fPeak)                 // frequency peak fpeak is globally defined
   {
      s=9.77*pow(f/fPeak,-2.5);
   } else
   {
      s=6.97*pow(f/fPeak,5.0);
   }
   D=Gamma(s+1)*pow(ang,s)/(2*sqrt(Pi)*Gamma(s+0.5));
   return(D);
}
//---------------------------------------------------------------------------
double EPM(double f)
{
   double EPM;    // EPM=Energy according to the Pierson-Moskowitz formula
   EPM=alpha*g*g*exp(-1.25*pow(f/fPeak,-4))/pow(2*Pi,4);
   return EPM;
}
//---------------------------------------------------------------------------
double EnergyDistribution(double f, double theta)
{
   return EPM(f)*MitsuyasuDistribution(f,theta);
}
//---------------------------------------------------------------------------
double Amplitude(double f, double theta, double k)
{
   double omega = 10.0;            // area of the water surface part
   double amp;                     // amplitude of the wave
   amp=sqrt(EnergyDistribution(f,theta)*g*Pi*Pi/(k*f*omega));
   return amp;
}
//---------------------------------------------------------------------------
void InitWater()
{
   fPeak=0.13*g/WindSpeed;         // calculate peak frequency
   WAVE *w = &wave[0];
   nWaves=0;                       // number of useable waves calculated so far 

   while (nWaves<maxwave)
   {
      float f = fPeak + rand() / 16384.0f * 0.5 - 0.5;

      if (f>0.0)
      {
          w->lambda      = 2.0*Pi/(pow( 2.0*Pi*f,2.0)/g);
          w->k           = 2.0*Pi/w->lambda;
          w->omega       = sqrt(g*w->k);
          w->freq        = w->omega / (2.0*Pi);
          w->periode     = 1.0f/w->freq;
          w->direction   = (rand()/16384.0f+0.5f)*Pi;
          w->dirX        = cos(w->direction)*0.5f;
          w->dirY        = sin(w->direction)*0.5f;

          double phi0    = rand()/16384.0f*Pi;
          double a0      = Amplitude(w->freq,w->direction-WindDirection,w->k);

          w->amplitude   = a0*cos( phi0 );
          w->phase       = a0*sin( phi0 );

          if (fabs(w->amplitude)>=0.0001f )
          {
             w ++;
             nWaves ++;
          }
      }
   }

   for (int i=0; i<WaterX*WaterY; i++ )
   {
      WaterNormal[ i ].x = 0.0f;
      WaterNormal[ i ].y = 1.0f;
      WaterNormal[ i ].z = 0.0f;
   }
}
//---------------------------------------------------------------------------
#define fx(x) ((x<0)?(x+WaterX):((x>=WaterX)?(x-WaterX):x))
#define fy(y) ((y<0)?(y+WaterY):((y>=WaterY)?(y-WaterY):y))

void AnimateWater()
{
  float time=GetTickCount()/500.0;

  for(int j=0; j<WaterY; j++)
    for(int i=0; i<WaterX; i++)
    {
      double h=0.0f;
      for ( int k = 0; k<nWaves; k++)
      {
        h += wave[k].amplitude*
                sin(wave[k].phase+
                wave[k].omega*(time+
                wave[k].dirX*i+wave[k].dirY*j));
      }
      WaterHeight[i+j*WaterX]=h;
    }

  for (int j=0; j<WaterY; j++)
    for ( int i=0; i<WaterX; i++)
    {
       float x, y, z;

       x = WaterHeight[fx(i+1)+j*WaterX]-WaterHeight[fx(i-1)+j*WaterX];
       z = WaterHeight[i+fy(j+1)*WaterX]-WaterHeight[i+fy(j-1)*WaterX];
       y = 5.0;

       WaterNormal[i+j*WaterX].x = x;
       WaterNormal[i+j*WaterX].y = y;
       WaterNormal[i+j*WaterX].z = z;
    }
}
//---------------------------------------------------------------------------
__fastcall TFormMain::TFormMain(TComponent* Owner)
    : TForm(Owner)
{
    Application->OnIdle = IdleLoop;   // Define idle function
    size = 100.0f;
}
//---------------------------------------------------------------------------
void __fastcall TFormMain::IdleLoop(TObject*, bool& done)
{
    done = false;               // request more idle time from the system
    RenderGLScene();
    SwapBuffers(hdc);
}
//---------------------------------------------------------------------------
void __fastcall TFormMain::FormCreate(TObject *Sender)
{
    hdc = GetDC(Handle);           // get device context from main window
    SetPixelFormatDescriptor();
    hrc = wglCreateContext(hdc);   // use device context to create
                                   // a rendering context
    if(hrc == NULL)
    	ShowMessage("Creating Rendering Context failed.");
    wglMakeCurrent(hdc,hrc);       // tell windows to use hdc and rdc
    if(wglMakeCurrent(hdc, hrc) == false)
    	ShowMessage("MakeCurrent of rendering context failed.");
    w = ClientWidth;               // width of main windows client area
    h = ClientHeight;              // height of main windows client area

    glEnable(GL_DEPTH_TEST);

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);  // clear background to black

    // initialize the ocean surface
    InitWater();
}
//---------------------------------------------------------------------------
void __fastcall TFormMain::SetPixelFormatDescriptor()
{
    PIXELFORMATDESCRIPTOR pfd = {
    	sizeof(PIXELFORMATDESCRIPTOR),       // size of the structure
        1,                                   // structure version
        PFD_DRAW_TO_WINDOW |                 // draw directly to window
                                             // (not to a bitmap file)
        PFD_SUPPORT_OPENGL |                 // allow DC to support opengl calls
        PFD_DOUBLEBUFFER,                    // use double buffer
        PFD_TYPE_RGBA,                       // use RGBA color mode (A = alpha)
        24,                                  // use 24 bit color
        0,0,0,0,0,0,0,0,0,0,0,0,0,           // not (yet) used
        32,                                  // use 32 bit z buffer
        0,0,                                 // not (yet) used
        PFD_MAIN_PLANE,                      // draw to main plane
        0,0,0,                               // not (yet) used
    };
    PixelFormat = ChoosePixelFormat(hdc, &pfd);  // choose and set the
    SetPixelFormat(hdc, PixelFormat, &pfd);      // appropriate pixelformat
}
//---------------------------------------------------------------------------
void __fastcall TFormMain::FormResize(TObject *Sender)
{

    GLfloat nRange = GridSize;

    w = ClientWidth;      // check new client width of the main window
    h = ClientHeight;     // check new client height of the main window

    if(h == 0)            // prevent a division by zero, by making sure that
    	h = 1;            // windows height is at least 1

    glViewport(0, 0, w, h);         // reset viewport

    glMatrixMode(GL_PROJECTION);    // add perspective to scene
    glLoadIdentity();               // restore matrix to original state

    gluLookAt( 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f);

    if (w <= h)
        glOrtho (-nRange, nRange, -nRange*h/w, nRange*h/w, -nRange, nRange);
    else
	glOrtho (-nRange*w/h, nRange*w/h, -nRange, nRange, -nRange, nRange);

    glMatrixMode(GL_MODELVIEW);

    glLoadIdentity();
}
//---------------------------------------------------------------------------
void __fastcall TFormMain::RenderGLScene()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    DrawObjects();
    glFlush();
}
//---------------------------------------------------------------------------
void __fastcall TFormMain::DrawObjects()
{
   // put your rendering functions here

   AnimateWater();

   glEnable(GL_NORMALIZE);
   glDepthFunc(GL_EQUAL);
   for(int j=0; j<WaterY; j++)
   {
      glBegin(GL_QUAD_STRIP);
      for(int i=0; i<WaterX; i++)
      {
         int x=i&(WaterX-1);
         int y=j&(WaterY-1);
         int y1=(j+1)&(WaterY-1);
         //glNormal3fv(&WaterNormal[x+y*WaterX]);
         glTexCoord2f(i*0.3f,j*0.3f);
         glVertex3f((i-WaterX/2),WaterHeight[x+y*WaterX],(j-WaterY/2));
         //glNormal3fv(&WaterNormal[x+y1*WaterX]);
         glTexCoord2f(i*0.3f,(j+1)*0.3f);
         glVertex3f((i-WaterX/2),WaterHeight[x+y1*WaterX],(j+1-WaterY/2));
      }
      glEnd();
   }

}
//---------------------------------------------------------------------------
void __fastcall TFormMain::FormPaint(TObject *Sender)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glFlush();
    DrawObjects();
}
//---------------------------------------------------------------------------
void __fastcall TFormMain::FormDestroy(TObject *Sender)
{
    wglMakeCurrent(NULL, NULL);
    wglDeleteContext(hrc);
}

